# -*- coding: utf-8 -*-

from ultron.tradingday.Enums._Months cimport Months as mt


cpdef enum Months:
    January = mt.January
    February = mt.February
    March = mt.March
    April = mt.April
    May = mt.May
    June = mt.June
    July = mt.July
    August = mt.August
    September = mt.September
    October = mt.October
    November = mt.November
    December = mt.December